#!/usr/local/bin/python

"""The Assignments client/server protocol.

This implementation of the Assignments protocol is synchronous. It would be
interesting to see if the asyncore module could be used to produce a portable
asynchronous implementation so that users could cancel requests in progress,
GUI updates could be processed during lengthy data transfers, default
socket-layer timeouts could be shortened, etc.

(In the Server class's _methodCall() method, I tried implementing a periodic
callback in the while 1: loop to get some asynchronous behavior. That worked
fine, but I found that for really big requests we were spending the most time
in httplib's getreply() method before even reaching my _methodCall() loop. So
in order to be able to give the user feedback uniformly throughout the long
wait, we'll need a more elaborate approach.)

This module has at least one other design issue that's worth further
consideration: The return value from an Assignments RPC is currently just the
return value from xmlrpclib, i.e. a data structure built directly out of
Python's primitive types (lists, dictionaries, strings, etc.) that directly
represents the XML-RPC method response parameters. This means that any code
calling an Assignments RPC must know all about the internal structure of the
responses; for example, a lot of code "knows" that result[0][0] is a hostname
and result[0][2] is a dictionary containing other host information. This in
turn means that anyone reading the code must be intimately familiar with the
details of the protocol in order to understand the mysterious manipulations of
data[-1][2]['IPv4Address'] and so forth. In order to make such code easier to
understand, a layer of abstraction should be added so that callers will instead
get an array of Host objects, MailExchanger objects, etc., which would have
well-defined methods for extracting data about the records they represent.

Note that most other modules will not use this module directly, and will
instead use a GUI-aware Server instance object from the AsgsServer module.

Classes:

Server -- Communication with the Assignments server.

Exception classes:

Error -- Base class for Assignments protocol errors.
UserAborted -- Operation aborted by user.
RPCError -- XML-RPC fault.
TransportError -- HTTP protocol error.
AuthenticationError -- HTTP Authentication error.
GSSAPIError -- gssapi-over-sasl-over-http error.
PermissionError -- user has no Assignments permissions
NetworkError -- Socket error.

"""

import base64
import logging
import os
import socket
import struct
import sys

try:  # py2
    import httplib
except ImportError:  # py3
    import http.client as httplib

try:  # py2
    import xmlrpclib
except ImportError:  # py3
    import xmlrpc.client as xmlrpclib

import kerberos

from . import Constants


# authentication states
_AUTH_IDLE = 'Auth-idle'
_AUTH_STARTED = 'Auth-started'
_AUTH_PENDING = 'Auth-pending'
_AUTH_DONE = 'Auth-done'


# Security Layers per RFC 4752, Section 3.3
GSSAPI_LAYER_NONE = 1
GSSAPI_LAYER_INTEG = 2
GSSAPI_LAYER_CONF = 4


class Error(Exception):

    """Base class for Assignments protocol errors.

    Public attributes:

    args -- Exception arguments (tuple).
    code -- Numeric error code.
    msg -- Descriptive error message.

    """

    def __init__(self, *args):
        self.args = args
        if len(args) > 1:
            self.code = args[0]
            self.msg = args[1]
        elif len(args) == 1:
            self.code = -1
            self.msg = args[0]
        else:
            self.code, self.msg = 0, 'unknown error'


class UserAborted(Error):
    """Operation aborted by user."""
    pass


class RPCError(Error):
    """XML-RPC fault."""
    pass


class TransportError(Error):
    """HTTP protocol error."""
    pass


class GSSAPIError(TransportError):
    """GSSAPI-over-http error"""
    pass


class AuthenticationError(TransportError):
    """HTTP authentication error."""
    pass


class NetworkError(Error):
    """Socket error."""
    pass


class RemoteProcedure:

    """Callable remote procedure objects.

    This class is based closely on the Method class from the xmlrpclib.py
    module, with one addition: It allows that caller to supply keyword
    arguments that will be passed on to the send function.

    """

    def __init__(self, send, name):
        """Constructor.

        Arguments:

        send -- A function used to send a remote procedure call to the server
                (given the method name as its first argument, followed by any
                other positional and keyword arguments provided by the caller).
        name -- The XML-RPC method name.

        """
        self.__send = send
        self.__name = name

    def __getattr__(self, name):
        """Look up an attribute name."""
        # Allows intelligent (nested) handling of RPC names with dots in them.
        return RemoteProcedure(self.__send, "%s.%s" % (self.__name, name))

    def __call__(self, *args, **keywords):
        """Call the remote procedure with specified arguments."""
        return self.__send(self.__name, args, **keywords)

    def __repr__(self):
        """Return a canonical string representation."""
        return ("<Assignments RPC for %s>" % self.__name)
    __str__ = __repr__

    def __nonzero__(self):
        """Return truth value."""
        return 1


class Server:

    """Communication with the Assignments server.

    Public methods:

    set_user_agent() -- Set the HTTP User-Agent: header.

    Other public attributes:

    The name of an Assignments remote procedure call may be looked up as an
    attribute of any AsgsProtocol.Server instance to obtain a callable remote
    procedure object for the corresponding XML-RPC method.

    This class is similar to the xmlrpclib.Server class from the xmlrpclib.py
    module, but it's specifically tailored for the Assignments application.
    It's not implemented as a subclass of xmlrpclib.Server since it would need
    to override all of the important methods anyway.

    Key functional differences from the xmlrpclib.Server class are:

    1. Assignments uses a fixed server host and responder path, so there's
       no need for the constructor to parse out this information from
       arbitrary URLs.

    2. Assignments uses gssapi-over-http for authentication.

       There are references to SASL, but since the scheme assumes the
       GSSAPI mechanism, we don't actually need all the extra SASL gunk.

    3. XML-RPC faults with fault codes less than zero may be sent by the
       Assignments server to represent warnings that can be overridden at
       the client's discretion. The Server._methodCall method can
       optionally use a callback to determine if a given warning should be
       overridden, and if so it will modify the request appropriately and
       resend it. The FaultDialog module can be used to present errors and
       warnings to the user and to ask the user if a warning should be
       overridden.

    4. A protocol ID string is prepended to all XML-RPC method call names.

    Other implementation notes:

    1. The name of every attribute of an AsgsProtocol.Server instance
       contains at least one underscore character (_). This convention
       ensures that none of the instance object's attributes will conflict
       with XML-RPC method call names since the Assignments protocol
       specification guarantees that those names will not contain
       underscores.

    2. This class was developed for use with version 0.9.5-990202 of the
       xmlrpclib.py module. Minor changes were made to make it work with
       version 0.9.8-990621, but the basic design was not revisited. Since
       the newer xmlrpclib code factors out the transport functions into a
       separate class, a more elegant implementation may now be possible.

    """

    def __init__(self, fault_handler=None, host=Constants.ServerHost,
                 path=Constants.ServerPath, protocol_id=Constants.ProtocolID,
                 logger=logging.getLogger(__name__),
                 authzname=''):
        """Constructor.

        Optional keyword arguments:

        fault_handler -- A callable object used to report errors or warnings.
                If it is not None (the default), it will be called with two
                arguments: A numeric fault code, and a human-readable error
                message. A fault code of zero indicates that the user aborted
                the operation. A fault code greater than zero indicates that
                an error occurred while performing the operation. A fault code
                less than zero represents a warning that may optionally be
                overridden. If the fault code is less than zero and it should
                be overridden, fault_handler must return 1; in all other cases,
                it must return 0.
        host -- Server hostname. Defaults to Constants.ServerHost.
        path -- XML-RPC responder path. Defaults to Constants.ServerPath.
        protocol_id -- Protocol ID string. Defaults to Constants.ProtocolID.

        """
        self._host = host
        self._path = path
        self._fault_ok = fault_handler
        self._auth_state = _AUTH_IDLE  # authentication state
        self._log = logger
        self._headers = self._fixedHeaders()
        self._protocol_id = "%s.%%s" % protocol_id
        self._http = None       # persistant HTTP connection
        self._have_authed = 0
        self.authzname = authzname  # proxy for admin

        # This dict is used to hook RPC calls to local callables instead.
        # This is generally done for debugging/development purposes,
        # such as when the server hasn't or can't be changed during
        # client change testing.
        self._intercept_and_fake = {}

    def setauthzname(self, authzname):
        self.authzname = authzname

    def __getattr__(self, attribute):
        """Return a callable object that can execute an XML-RPC method call."""

        # This bit checks to see if this call is being hooked locally
        # for implementation or extension.  This is intended for debugging
        # usage only - actual permanent changes in functionality should be
        # implemented on the server, of course.
        if attribute in self._intercept_and_fake:
            return self._intercept_and_fake[attribute]

        return RemoteProcedure(self._methodCall, self._protocol_id % attribute)

    def _methodCall(self, methodName, params, **faultdata):
        """Call an XML-RPC method and return the response parameter data.

        This method may raise an AsgsProtocol.NetworkError exception if it
        cannot talk to the server, or an AsgsProtocol.UserAborted exception
        if it catches a KeyboardInterrupt, or an AsgsProtocol.TransportError
        if it receives an HTTP status code other than 200 from the server, or
        an AsgsProtocol.RPCError exception if it receives a fault response from
        the server that is not overridden by the self._fault_ok function. The
        self._fault_ok function is always called before any of these exceptions
        is raised.

        Arguments:

        methodName -- The XML-RPC method name.
        params -- A tuple containing parameters for the XML-RPC method call.

        Optional keyword arguments may be provided, which will be passed to the
        fault handler function self._fault_ok in the event of an error.

        """
        fault_ok = self._fault_ok
        try:
            request = xmlrpclib.dumps(params, methodName) or ''
            self._log.debug("XML-RPC request: %s", request)
            # handle IDLE-TIMEOUT
            response = self._transport(request, **faultdata)
            parser, unmarshaler = xmlrpclib.getparser()
            response_read = response.read
            parser_feed = parser.feed
            while 1:

                data = response_read()

                if not data:
                    break
                self._log.debug("XML-RPC response: %s", data)
                parser_feed(data)
            parser.close()
            xml = unmarshaler.close()
            if len(xml) == 1:
                return xml[0]
            else:
                return xml
        except KeyboardInterrupt:
            err = 0, 'Aborting.'
            if fault_ok is not None:
                fault_ok(*err, **faultdata)
            raise UserAborted(err)
        except socket.error as error:
            if fault_ok is not None:
                message = ('The Assignments server could not be contacted '
                           'due to a network communications error: %s'
                           % error.args[-1])
                fault_ok(1, message, **faultdata)
            raise NetworkError(error.args, sys.exc_info()[2])
        except AuthenticationError as error:
            if fault_ok is not None:
                message = 'You do not have permissions to use Assignments.'
                fault_ok(1, message, **faultdata)
            raise AuthenticationError(error.args, sys.exc_info()[2])
        except kerberos.GSSError as error:
            if fault_ok is not None:
                message = self._ticket_warning(error.message)
                fault_ok(1, message, **faultdata)
            raise AuthenticationError(error, sys.exc_info()[2])
        except xmlrpclib.Fault as fault:
            err = fault.faultCode, fault.faultString
            if fault_ok is not None and fault_ok(*err, **faultdata):
                # The fault is okay: Add this fault code to our list of
                # overrides and resend the request.
                if (len(params) == 0
                        or not isinstance(params[-1], dict)
                        or len(params[-1]) != 1
                        or 'override' not in params[-1]):
                    # We don't have an override list yet, so we'll add one.
                    params = params + ({'override': []},)
                params[-1]['override'].append(fault.faultCode)
                return self._methodCall(methodName, params, **faultdata)
            else:
                raise RPCError(err)
        except MemoryError:
            self._reset_connection()
            if fault_ok is not None:
                message = ('The Assignments server could not be contacted '
                           'due to a memory error.  Please try altering your '
                           'request to limit the size.')
                fault_ok(1, message, **faultdata)
            raise NetworkError

    def _reset_connection(self):
        """set up our HTTP connection and authN variables"""
        self._auth_state = _AUTH_IDLE
        self._http = None
        self._num_reconnect += 1

    def _transport(self, body, **faultdata):
        """Send an HTTP request and return the response's entity body.

        Arguments:

        body -- The entity body for the HTTP request.

        Optional keyword arguments may be provided, which will be passed to
        the fault handler function self._fault_ok in the event of a transport
        error.

        """
        num_lost_auth = 0
        self._num_reconnect = 1
        server_creds = ''
        while 1:
            try:
                self._log.debug("AUTH STATE: %s", self._auth_state)
                if num_lost_auth > 1:
                    if self._fault_ok is not None:
                        message = 'Server error: could not authenticate'
                        message += ' with the server.'
                        self._fault_ok(10, message, **faultdata)
                        raise TransportError(message)
                if self._num_reconnect > 2:
                    if self._fault_ok is not None:
                        message = ('Assignments is having trouble talking '
                                   'to its server.')
                        self._fault_ok(10, message, **faultdata)
                        raise TransportError(message)
                if not self._http:
                    self._log.debug('creating http connection')
                    self._http = httplib.HTTPConnection(self._host)
                headers = self._headers + [('Content-Length', str(len(body)))]
                if self._auth_state == _AUTH_DONE:
                    auth_header = 'SASL id="%s"' % self._authid
                    self._have_authed = 1
                elif self._auth_state == _AUTH_IDLE:
                    self._authid = None
                    # see RFC-4752, Section 3.1
                    result, self._gssctx = kerberos.authGSSClientInit(
                            service="{}@{}".format(
                                    Constants.Service,
                                    Constants.Server
                            ),
                            gssflags=(kerberos.GSS_C_MUTUAL_FLAG |
                                      kerberos.GSS_C_INTEG_FLAG |
                                      kerberos.GSS_C_SEQUENCE_FLAG),
                            mech_oid=kerberos.GSS_MECH_OID_KRB5,
                    )
                    if result != kerberos.AUTH_GSS_COMPLETE:
                        raise GSSAPIError("GSSAPI client initializtion")
                    done = kerberos.authGSSClientStep(self._gssctx, "")
                    creds = kerberos.authGSSClientResponse(self._gssctx)
                    auth_header = ('SASL mechanism="%s",credentials=%s'
                                   % (Constants.Mechanism, creds))
                elif self._auth_state == _AUTH_STARTED:
                    self._log.debug("Client Step")
                    done = kerberos.authGSSClientStep(
                            self._gssctx, server_creds
                    )
                    self._log.debug("Client Response")
                    creds = kerberos.authGSSClientResponse(self._gssctx)
                    if not creds:
                        creds = ''
                    self._log.debug("Creds: %s", creds)
                    auth_header = ('SASL id="%s",credentials=%s'
                                   % (self._authid, creds))
                    if done == kerberos.AUTH_GSS_COMPLETE:
                        self._auth_state = _AUTH_PENDING
                elif self._auth_state == _AUTH_PENDING:
                    self._log.debug("Client Unwrap")

                    # unwrap the server's stuff and examine it
                    done = kerberos.authGSSClientUnwrap(self._gssctx,
                                                        server_creds)
                    token = kerberos.authGSSClientResponse(self._gssctx)
                    self._log.debug("Token: %s", token)
                    data = base64.b64decode(token)
                    if len(data) != 4:
                        raise GSSAPIError("Server sent invalid token")

                    layer_mask, a, b, c = struct.unpack('BBBB', data)

                    if layer_mask & GSSAPI_LAYER_NONE:
                        self._log.debug("Security layer: NONE")
                    if layer_mask & GSSAPI_LAYER_INTEG:
                        self._log.debug("Security layer: INTEGRITY")
                    if layer_mask & GSSAPI_LAYER_CONF:
                        self._log.debug("Security layer: CONFIDENTIALITY")

                    msg_size = (a << 16) | (b << 8) | c
                    self._log.debug("Max message size: %d", msg_size)

                    # construct my stuff and wrap it
                    layer_mask = GSSAPI_LAYER_NONE
                    msg_size = 0  # as required for NONE
                    a = (msg_size >> 16) & 0xff
                    b = (msg_size >> 8) & 0xff
                    c = msg_size & 0xff
                    data = struct.pack("BBBB", layer_mask, a, b, c)
                    token = base64.b64encode(data).decode('ascii')
                    done = kerberos.authGSSClientWrap(self._gssctx,
                                                      token, self.authzname)
                    creds = kerberos.authGSSClientResponse(self._gssctx)

                    auth_header = ('SASL id="%s",credentials=%s'
                                   % (self._authid, creds))

                self._authorization = [('Authorization', auth_header)]
                # Send the request here
                try:
                    self._http.putrequest("POST", self._path)
                except httplib.CannotSendRequest:
                    self._log.debug('Lost connection to server on attempt %d',
                                    self._num_reconnect)
                    self._reset_connection()
                    continue
                for header in headers + self._authorization:
                    self._log.debug("%s: %s", *header)
                    self._http.putheader(*header)
                try:
                    self._http.endheaders()
                    self._http.send(body.encode('utf8'))
                except socket.error as error:
                    self._log.debug('Socket error (%s). Retries: %d',
                                    error.args[-1], self._num_reconnect)
                    self._reset_connection()
                    continue
                # Give server 75 seconds to send a response line
                self._http.sock.settimeout(75)
                try:
                    server_response = self._http.getresponse()
                except (httplib.ResponseNotReady, httplib.BadStatusLine):
                    self._log.debug('Server closed connection '
                                    '(ResponseNotReady) '
                                    'on attempt %d', self._num_reconnect)
                    self._reset_connection()
                    continue
                if self._http.sock:
                    # httplib may close the socket if server requests it
                    # i.e. via the 'Connection: close' header.
                    self._http.sock.settimeout(None)
                status = server_response.status
                msg = server_response.reason
                server_headers = server_response.msg
                self._log.debug("%s %s", status, msg)
                self._log.debug("%s", server_headers)
                if not server_headers:
                    self._log.debug('Lost connection to server. '
                                    'Reconnect try %d',
                                    self._num_reconnect)
                    self._reset_connection()
                    continue
                if status == 403:
                    raise AuthenticationError(status, msg)
                try:
                    server_auth_header = server_response.getheader(
                            'WWW-Authenticate'
                    )
                except KeyError:
                    raise GSSAPIError('Server sent no WWW-Authenticate')
                if not server_auth_header:
                    self._log.debug('Lost connection to server. '
                                    'Reconnect try %d',
                                    self._num_reconnect)
                    self._reset_connection()
                    continue
                if server_auth_header.startswith('SASL mechanism="'):
                    # server has reset authentication
                    raise GSSAPIError('Server reset authentication.')
                # authentication is done if server says 200
                if status == 200:
                    if server_auth_header != 'SASL id="%s"' % self._authid:
                        raise GSSAPIError('server sent wrong auth id')
                    self._auth_state = _AUTH_DONE
                    return server_response
                server_response.read()
                # anything other than a 401 response is bad news
                if status != 401:
                    self._auth_state = _AUTH_IDLE
                    if self._fault_ok is not None:
                        self._log.debug("Transport Failure %d: %s",
                                        status, msg)
                        self._fault_ok(10, message, **faultdata)
                        raise TransportError(status, msg)
                if self._auth_state == _AUTH_DONE:
                    raise GSSAPIError('Server un-authenticated us.')
                else:
                    # status == 401
                    if self._auth_state == _AUTH_IDLE:
                        self._auth_state = _AUTH_STARTED
                    (scheme, other) = server_auth_header.split()
                    if scheme != 'SASL':
                        self._log.debug('Server sent bad auth scheme.')
                        self._auth_state = _AUTH_IDLE
                    (sid, challenge) = other.split(',')
                    try:
                        idx = challenge.index('=') + 1
                        server_creds = challenge[idx:]
                    except:
                        raise GSSAPIError('Server creds badly encoded')
                    try:
                        sid = sid.split('"')[1]
                    except:
                        err = 'Server sent badly formed auth header (id)'
                        raise GSSAPIError(err)
                    if self._authid:
                        if sid != self._authid:
                            err = ('id mismatch. server: %s, us: %s\n'
                                   % (sid, self._authid))
                            raise GSSAPIError(err)
                    else:
                        self._authid = sid
            except GSSAPIError as msg:
                self._auth_state = _AUTH_IDLE
                num_lost_auth += 1
                self._log.debug(msg)
                continue
        # end while 1

    def _ticket_warning(self, err):
        """Create warning message when there is a kerberos error.

        Arguments:

        err -- The error message as returned from the kerberos module.
        """
        if err.endswith('(Ticket expired)'):
            message = ('Your Kerberos tickets have expired. Please renew'
                       ' them and try again.')
        elif err.endswith('(No credentials cache found)'):
            if Constants.os_name == 'nt':
                if self._have_authed:
                    message = ("Assignments can't read your Kerberos tickets. "
                               "You will need to restart Assignments to "
                               "correct this error.")
                else:
                    message = ('Please obtain Kerberos tickets before '
                               'starting Assignments.')
            else:
                message = 'Please obtain Kerberos tickets and try again.'
        else:
            message = ('There has been an error using your Kerberos '
                       'tickets.  The error was %s' % err)
        return message

    def set_user_agent(self, user_agent):
        """Set the HTTP User-Agent: header.

        Arguments:

        user_agent -- A string of product tokens for HTTP's User-Agent: header.

        """
        self._headers = self._fixedHeaders(user_agent)

    def _fixedHeaders(self, user_agent=None):
        """Set values for the set of HTTP headers that won't change.

        Optional keyword arguments:

        user_agent -- A string of product tokens for HTTP's User-Agent: header.

        """
        return [
            ('User-Agent', user_agent or "asgsproto/%s" % Constants.Version),
            ('Host', self._host),
            ('Content-Type', 'text/xml'),
            ('Connection', 'keep-alive')
        ]

    def __repr__(self):
        """Return a canonical string representation."""
        return ("<Assignments server http://%s%s>" % (self._host, self._path))
    __str__ = __repr__

    def __nonzero__(self):
        """Return truth value."""
        return 1
